import numpy as np
import pytz
from astropy.io import fits
from astropy.coordinates import ICRS
from astropy.time import Time
from datetime import datetime
from reproject import reproject_interp
from reproject.mosaicking import find_optimal_celestial_wcs

__all__ = ['sanitize_image']


def sanitize_image(image, output_file, overwrite=False):
    """
    Transform a FITS image so that it is in equatorial coordinates with a TAN
    projection and floating-point values, all of which are required to work
    correctly in WWT at the moment.

    Image can be a filename, an HDU, or a tuple of (array, WCS).
    """

    # The reproject package understands the different inputs to this function
    # so we can just transparently pass it through.

    wcs, shape_out = find_optimal_celestial_wcs([image], frame=ICRS(),
                                                projection='TAN')

    array, footprint = reproject_interp(image, wcs, shape_out=shape_out)

    fits.writeto(output_file, array.astype(np.float32),
                 wcs.to_header(), overwrite=overwrite)


def validate_traits(cls, traits):
    '''
    Helper function to ensure user-provided trait names match those of the
    class they're being used to instantiate.
    '''
    mismatch = [key for key in traits if key not in cls.trait_names()]
    if mismatch:
        raise KeyError('Key{0} {1} do{2}n\'t match any layer trait name'
                       .format('s' if len(mismatch) > 1 else '',
                               mismatch,
                               '' if len(mismatch) > 1 else 'es'))


def ensure_utc(tm, str_allowed):
    '''
    Helper function to convert a time object (Time, datetime, or UTC string
    if str_allowed == True) into UTC before passing it to WWT.
    str_allowed is True for wwt.set_current_time (core.py) and False for TableLayer's 'time_att' implementation (layers.py).
    '''

    if tm is None:
        utc_tm = datetime.utcnow().astimezone(pytz.UTC).isoformat()

    elif isinstance(tm, datetime):
        if tm.tzinfo is None:
            utc_tm = pytz.utc.localize(tm).isoformat()
        elif tm.tzinfo == pytz.UTC:
            utc_tm = tm.isoformat()
        else:  # has a non-UTC time zone
            utc_tm = tm.astimezone(pytz.UTC).isoformat()

    elif isinstance(tm, Time):
        utc_tm = tm.to_datetime(pytz.UTC).isoformat()

    else:
        if str_allowed:  # is an ISOT string
            dt = Time(tm, format='isot').to_datetime(pytz.UTC)
            utc_tm = dt.isoformat()
        else:
            raise ValueError('Time must be a datetime or astropy.Time object')

    return utc_tm



# Copyright 2018-2021 the .NET Foundation
# Licensed under the BSD license

"""
Some basic tests.
"""

import pytest
from traitlets import TraitError

from ..core import BaseWWTWidget


def test_color_validation():
    widget = BaseWWTWidget()

    widget.constellation_figure_color = 'red'
    assert widget.constellation_figure_color == '#ff0000'

    widget.constellation_figure_color = '#ff0000'
    assert widget.constellation_figure_color == '#ff0000'

    widget.constellation_figure_color = (1, 0, 0)
    assert widget.constellation_figure_color == '#ff0000'

    with pytest.raises(TraitError) as exc:
        widget.constellation_figure_color = (1, 0, 0, 2)
    assert exc.value.args[0] == ('This attribute\'s color must be a string '
                                 '(a recognized matplotlib color name or hex '
                                 'code) or an RGB tuple of 3 floats')



# This file contains the definition of the Python part of the WWT Jupyter
# widget. Note that we don't tag each trait from BaseWWTWidget as sync=True
# because we instead use JSON messages to transmit any changes between the
# Python and Javascript parts so that we can re-use this for the Qt client.

import ipywidgets as widgets
import numpy as np
from traitlets import Unicode, default, link, directional_link

from ipyevents import Event as DOMListener
from ipykernel.comm import Comm

from .core import BaseWWTWidget
from .layers import ImageLayer
from .logger import logger
from .jupyter_server import serve_file

__all__ = ['WWTJupyterWidget', 'WWTLabApplication', 'connect_to_app']

_npm_version = '^1.0.0'  # cranko internal-req npm:pywwt
VIEW_MODULE_VERSION = _npm_version
MODEL_MODULE_VERSION = _npm_version

R2D = 180 / np.pi
R2H = 12 / np.pi

dom_listener = DOMListener()


@widgets.register
class WWTJupyterWidget(widgets.DOMWidget, BaseWWTWidget):
    """
    An AAS WorldWide Telescope Jupyter widget.

    Parameters
    ----------

    hide_all_chrome : optional `bool`
        Configures the WWT frontend to hide all user-interface "chrome".

    """

    _view_name = Unicode('WWTView').tag(sync=True)
    _model_name = Unicode('WWTModel').tag(sync=True)
    _view_module = Unicode('pywwt').tag(sync=True)
    _model_module = Unicode('pywwt').tag(sync=True)
    _view_module_version = Unicode(VIEW_MODULE_VERSION).tag(sync=True)
    _model_module_version = Unicode(MODEL_MODULE_VERSION).tag(sync=True)

    _appUrl = Unicode('').tag(sync=True)

    def __init__(self, hide_all_chrome=False):
        # In the future we might want to make it possible to use the WWT-hosted
        # app instead of the bundled version.
        #
        # The JS frontend will automagically prepend the Jupyter base URL if
        # needed and turn this into an absolute URL.

        _maybe_perpetrate_mega_kernel_hack()

        self._appUrl = '/wwt/research/'

        widgets.DOMWidget.__init__(self)
        dom_listener.source = self
        dom_listener.prevent_default_action = True
        dom_listener.watched_events = ['wheel']

        self._controls = None

        self.on_msg(self._on_ipywidgets_message)

        BaseWWTWidget.__init__(self, hide_all_chrome=hide_all_chrome)

    def _on_ipywidgets_message(self, widget, content, buffers):
        """
        Called when we receive a "custom" ipywidgets message.

        NOTE: because this code is run asynchronously in Jupyter's comms
        architecture, exceptions and printouts don't get reported to the user --
        they just disappear. I don't know if there's a "right" way to address
        that.
        """

        # Special message from the ipywidgets bridge to indicate
        # when the first widget view is ready to accept messages.
        if content.get('type') == 'wwt_jupyter_widget_status':
            self._on_app_status_change(alive=content['alive'])

        self._on_app_message_received(content)

    def _actually_send_msg(self, payload):
        """
        Send a message to the app. In ipywidgets this is easy.
        """
        self.send(payload)

    @default('layout')
    def _default_layout(self):
        return widgets.Layout(height='400px', align_self='stretch')

    def _serve_file(self, filename, extension=''):
        return serve_file(filename, extension=extension)

    def _create_image_layer(self, **kwargs):
        """Returns a specialized subclass of ImageLayer that has some extra hooks for
        creating UI control points.

        """
        return JupyterImageLayer(parent=self, **kwargs)

    @property
    def layer_controls(self):
        if self._controls is None:
            opacity_slider = widgets.FloatSlider(value=self.foreground_opacity,
                                                 min=0, max=1, readout=False)
            foreground_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.foreground)
            background_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.background)
            link((opacity_slider, 'value'), (self, 'foreground_opacity'))
            link((foreground_menu, 'value'), (self, 'foreground'))
            link((background_menu, 'value'), (self, 'background'))
            self._controls = widgets.HBox([background_menu, opacity_slider, foreground_menu])
        return self._controls


class JupyterImageLayer(ImageLayer):
    def __init__(self, **kwargs):
        self._controls = None
        super(JupyterImageLayer, self).__init__(**kwargs)

    @property
    def controls(self):
        from .layers import VALID_STRETCHES, UI_COLORMAPS

        if self._controls is not None:
            return self._controls

        opacity = widgets.FloatSlider(
            description='Opacity:',
            value=self.opacity,
            min=0,
            max=1,
            readout=False,
            step=0.01,
            layout={'width': '200px'}
        )
        link((self, 'opacity'), (opacity, 'value'))

        stretch = widgets.Dropdown(
            description='Stretch:',
            options=VALID_STRETCHES,
            value=self.stretch,
            layout={'width': '200px'}
        )
        link((self, 'stretch'), (stretch, 'value'))

        # NB, this will crash if `self.cmap` is not one of our allowed values
        reverse_ui_colormaps = dict((kv[1], kv[0]) for kv in UI_COLORMAPS.items())
        colormap = widgets.Dropdown(
            description='Colormap:',
            options=UI_COLORMAPS.keys(),
            value=reverse_ui_colormaps[self.cmap.name],
            layout={'width': '200px'}
        )
        directional_link((colormap, 'label'), (self, 'cmap'), lambda x: UI_COLORMAPS[x])
        directional_link((self, 'cmap'), (colormap, 'label'), lambda x: reverse_ui_colormaps[x.name])

        vrange = widgets.FloatRangeSlider(
            description='Fine min/max:',
            value=[self.vmin, self.vmax],
            min=self._data_min,
            max=self._data_max,
            readout=True,
            layout={'width': '600px'},
            step=(self.vmax - self.vmin) / 100,
            format='.3g'
        )

        # Linkage must be manual since vrange uses a pair of values whereas we
        # have two separate traitlets.
        vrange.observe(self._vrange_slider_updated, names=['value'])

        def update_vrange(change):
            # Note: when this function is called, these values are indeed updated.
            vrange.value = (self.vmin, self.vmax)

        self.observe(update_vrange, names=['vmin', 'vmax'])

        def update_step(change):
            vrange.step = (vrange.max - vrange.min) / 100

        vrange.observe(update_step, names=['min', 'max'])

        coarse_min = widgets.FloatText(
            description='Coarse min:',
            value=self._data_min,
            layout={'width': '300px'}
        )
        link((coarse_min, 'value'), (vrange, 'min'))

        coarse_max = widgets.FloatText(
            description='Coarse max:',
            value=self._data_max,
            layout={'width': '300px'}
        )
        link((coarse_max, 'value'), (vrange, 'max'))

        self._controls = widgets.VBox([
            widgets.HBox([colormap, stretch, opacity]),
            widgets.HBox([coarse_min, coarse_max]),
            vrange,
        ])
        return self._controls

    def _vrange_slider_updated(self, change):
        self.vmin, self.vmax = change['new']


class WWTLabApplication(BaseWWTWidget):
    """
    A handle the WWT JupyterLab application.

    While other parts of pywwt create "widgets", bound to variables running
    inside Python notebooks, this class represents a connection to the
    standalone "application", which exists in JupyterLab independently of any
    one specific notebook. The Python API is the same, it's just that the JSON
    messages we send are routed to the separate application rather than our own
    iframe.
    """

    _comm = None
    _controls = None

    def __init__(self):
        _maybe_perpetrate_mega_kernel_hack()

        self._comm = Comm(target_name='@wwtelescope/jupyterlab:research', data={})
        self._comm.on_msg(self._on_comm_message_received)
        self._comm.open()

        super(WWTLabApplication, self).__init__()

    def _on_comm_message_received(self, msg):
        """
        Called when we receive a comms message.

        NOTE: because this code is run asynchronously in Jupyter's comms
        architecture, exceptions and printouts don't get reported to the user --
        they just disappear. I don't know if there's a "right" way to address
        that.
        """
        payload = msg['content']['data']

        # Special message from the hub indicating app liveness status
        if payload.get('type') == 'wwt_jupyter_viewer_status':
            self._on_app_status_change(alive=payload['alive'])
            # don't return -- maybe someone downstream can use this, and message
            # processing needs to handle all sorts of unexpected messages anyway

        self._on_app_message_received(payload)

    def _actually_send_msg(self, payload):
        self._comm.send(payload)

    def _serve_file(self, filename, extension=''):
        return serve_file(filename, extension=extension)

    def _create_image_layer(self, **kwargs):
        """Returns a specialized subclass of ImageLayer that has some extra hooks for
        creating UI control points.

        """
        return JupyterImageLayer(parent=self, **kwargs)

    @property
    def layer_controls(self):
        if self._controls is None:
            opacity_slider = widgets.FloatSlider(value=self.foreground_opacity,
                                                 min=0, max=1, readout=False)
            foreground_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.foreground)
            background_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.background)
            link((opacity_slider, 'value'), (self, 'foreground_opacity'))
            link((foreground_menu, 'value'), (self, 'foreground'))
            link((background_menu, 'value'), (self, 'background'))
            self._controls = widgets.HBox([background_menu, opacity_slider, foreground_menu])
        return self._controls


def connect_to_app():
    """
    Connect to a WWT application running inside a JupyterLab computational
    environment.

    For the time being, you must have opened the AAS WorldWide Telescope app
    inside JupyterLab. You can do this by clicking the large WWT icon in the
    JupyterLab launcher, or by invoking the "AAS WorldWide Telescope" command.
    You can open the JupyterLab command palette by typing
    Control/Command-Shift-C.

    Returns
    -------
    app : :class:`~pywwt.jupyter.WWTLabApplication`
        A connection to the WWT application running in JupyterLab.

    """
    # This function just exists because it seems nicer from a UX standpoint to
    # have the user call a function with this name, than to create a "connection
    # object".
    return WWTLabApplication()


def _maybe_perpetrate_mega_kernel_hack():
    """
    OK. So.

    The Python code running in this process is communicating with the WWT
    JavaScript frontend, which is running in another process that is possibly on
    the other side of the internet. Various WWT operations such as loading
    imagery need that frontend to perform web requests and the like. So much of
    what we want to do is just inherently asynchronous.

    Which is OK! We require sufficiently new Python that we can rely on
    async/await being available, and the Jupyter infrastructure is extremely
    async-friendly.

    Except that we have a problem. When our process is running inside Jupyter,
    all code executions are inherently triggered by "shell messages" telling the
    kernel to run some code. Shell messages are also used for Jupyter's "comms"
    infrastructure, which is how kernels get updates from the WWT frontend(s)
    about things that are going on. The problem is that the Python Jupyter
    kernels only process shell messages *sequentially*: while message evaluation
    can be asynchronous, they can't be evaluated simultaneously. So if we've got
    asynchronous user code that's attempting to do stuff with pywwt, we can't
    receive any updates from WWT while that code is running. This fundamentally
    breaks our ability to allow users to write code that interacts with WWT
    asynchronously.

    Our mega-hack solution is to allow some shell messages to be marked for
    "expedited" processing. We hack the kernel so that such messages can be
    processed even while another shell message -- such as an execution request
    -- is being dealt with. This breaks the logjam.

    Expedited messages have the following structure:

    {
        'content': {
            '_pywwtExpedite': true
        }
    }

    This allows ipywidgets custom messages to be marked for expedited
    processing, which we need for our ipywidget support.

    To amp up the debugging spew so that you can see what's going on inside
    the kernel (which is captured in the Jupyter Server output), run:

    ```
    import ipykernel.kernelbase
    ipykernel.kernelbase.Kernel.instance().log.setLevel('DEBUG')
    ```

    Please forgive me.
    """

    try:
        _maybe_perpetrate_mega_kernel_hack_inner()
    except Exception:
        logger.exception('failed to set up Jupyter kernel async hack')


def _maybe_perpetrate_mega_kernel_hack_inner():
    import asyncio
    import inspect
    import ipykernel.kernelbase

    kernel = ipykernel.kernelbase.Kernel.instance()
    orig_schedule_dispatch = kernel.schedule_dispatch

    if getattr(kernel, '_pywwt_mega_hack_installed', False):
        return

    # If asyncio doesn't think that there's a running event loop, we don't seem
    # to be running as a client of a full-on Jupyter server. In which case, play
    # it safe and don't do anything.

    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return

    # OK, it looks like we should try to do this ...

    async def dispatch_one_expedited_shell_message(idents, msg):
        """
        A bastardized version of `kernel.dispatch_shell`. Our arguments will be
        the result of `kernel.session.feed_identities`.

        Expedited shell messages have to be processed as straightforwardly as
        possible since their processing can happen while other shell message
        processing is happening. The kernel code is *not* build to handle the
        recursive-y, race-y things that can happen in this situation. So we
        avoid all function calls that we can possible get away with skipping.
        No set_parent, no should_handle, etc.
        """

        try:
            msg = kernel.session.deserialize(msg, content=True, copy=False)
        except Exception:
            kernel.log.error("Invalid Expedited Message (pywwt)", exc_info=True)
            return

        msg_type = msg['header']['msg_type']
        handler = kernel.shell_handlers.get(msg_type, None)

        if handler is None:
            kernel.log.warning("Unknown expedited message type (pywwt): %r", msg_type)
        else:
            kernel.log.debug("expedited (pywwt) %s: %s", msg_type, msg)

            try:
                result = handler(kernel.shell_stream, idents, msg)
                if inspect.isawaitable(result):
                    kernel.log.error("Expedited message (pywwt) produce an awaitable result")
            except Exception:
                kernel.log.error("Exception in expedited message handler (pywwt):", exc_info=True)

    def pywwt_schedule_shell_dispatch_with_expedite(*args):
        """
        A replacement of `kernel.schedule_dispatch` for shell messages.

        This peeks inside the message and triggers expedited processing if
        called for. If there are any issues, we fall back to regular processing.

        """
        expedited_it = False

        try:
            (msg,) = args
            idents, msg = kernel.session.feed_identities(msg, copy=False)
            # We can't deserialize() here: each message can only be deserialized once
            # due to the replay prevention framework.
            peek_content = kernel.session.unpack(msg[4].bytes)
            expedite_flag = peek_content.get('data', {}).get('content', {}).get('_pywwtExpedite')

            if expedite_flag:
                kernel.io_loop.add_callback(dispatch_one_expedited_shell_message, idents, msg)
                expedited_it = True
        finally:
            if not expedited_it:
                orig_schedule_dispatch(kernel.dispatch_shell, *args)

    kernel.shell_stream.on_recv(pywwt_schedule_shell_dispatch_with_expedite, copy=False)
    kernel._pywwt_mega_hack_installed = True
    logger.debug('installed Jupyter kernel message expedite hack')



# Copyright 2018-2021 the .NET Foundation
# Licensed under the BSD license

import os
from datetime import datetime

import pytest

try:
    import qtpy
except ImportError:
    QT_INSTALLED = False
except Exception:  # will catch PythonQtError
    QT_INSTALLED = False
else:
    QT_INSTALLED = True

try:
    import OpenGL.GL as gl
except ImportError:
    OPENGL_INSTALLED = False
else:
    OPENGL_INSTALLED = True

RUNNING_ON_CI = os.environ.get('CI') or os.environ.get('AGENT_OS')

if QT_INSTALLED and OPENGL_INSTALLED:
    from qtpy.QtWidgets import QOpenGLWidget

    class OpenGLWidget(QOpenGLWidget):
        def getOpenGLInfo(self):
            gl_parameters = {
                'vendor': gl.GL_VENDOR,
                'renderer': gl.GL_RENDERER,
                'version': gl.GL_VERSION,
                'shader': gl.GL_SHADING_LANGUAGE_VERSION
            }

            info = {}

            for key, setting in gl_parameters.items():
                try:
                    info[key] = gl.glGetString(setting).decode('ascii')
                except Exception:
                    info[key] = 'could not be determined'

            return info


def pytest_report_header(config):
    lines = []

    if QT_INSTALLED:
        # Get Python Qt bindings version
        lines.append("PyQt: {0}".format(qtpy.PYQT_VERSION))
        lines.append("PySide: {0}".format(qtpy.PYSIDE_VERSION))

        # Get WebEngine/WebKit info
        from qtpy.QtWebEngineWidgets import WEBENGINE
        lines.append("Web Framework: {0}".format('WebEngine' if WEBENGINE else 'WebKit'))

        if OPENGL_INSTALLED:
            # Get OpenGL version
            from .app import get_qapp
            get_qapp()
            widget = OpenGLWidget()
            widget.show()
            opengl_info = widget.getOpenGLInfo()
            widget.close()

            lines.append("OpenGL Vendor: {0}".format(opengl_info['vendor']))
            lines.append("OpenGL Renderer: {0}".format(opengl_info['renderer']))
            lines.append("OpenGL Version: {0}".format(opengl_info['version']))
            lines.append("Shader Version: {0}".format(opengl_info['shader']))
        else:
            lines.append("Could not determine OpenGL version (OpenGL package required)")
    else:
        lines.append("Qt not installed")

    return os.linesep + os.linesep.join(lines) + os.linesep


def pytest_unconfigure(config):
    if QT_INSTALLED:
        from .app import cleanup_qapp
        cleanup_qapp()


REFERENCE_TIME = datetime(2017, 2, 1, 0, 0, 0, 0)

if QT_INSTALLED:
    def _check_app_available():
        import os.path
        f = os.path.join(os.path.dirname(__file__), 'web_static', 'research', 'index.html')
        if not os.path.exists(f):
            raise Exception('You must run `python setup.py build` before running these tests (to get research app files)')

    @pytest.fixture(scope='function')
    def wwt_qt_client_isolated():
        _check_app_available()
        from . import qt

        # As mentioned in wait_for_test() in tests/__init__.py, on the CI
        # system, the macOS tests use software rendering and it is slooooooow.
        # We need to give a very generous deadline for checking the app's
        # aliveness since the renders can cause huge amounts of wall-clock time
        # to elapse.
        qt.APP_LIVENESS_DEADLINE = 45

        wwt = qt.WWTQtClient(block_until_ready=True, size=(400, 400))
        wwt.set_current_time(REFERENCE_TIME)
        wwt.pause_time()
        yield wwt
        wwt.close()

    wwt_qt_client = wwt_qt_client_isolated



# Copyright 2018-2021 the .NET Foundation
# Licensed under the BSD license

"""
This file defines the WWT Qt widget.

Most of the code here deals with differences between WebEngine and WebKit
(either of which may be available in Qt) and also deals with figuring out how we
know once WWT is set up.
"""

import json
import time

from qtpy.QtWebEngineWidgets import QWebEngineView, QWebEnginePage, WEBENGINE
from qtpy import QtWidgets, QtGui, QtCore

from .app import get_qapp
from .core import BaseWWTWidget
from .logger import logger
from .data_server import get_data_server

__all__ = ['WWTQtClient']

APP_LIVENESS_DEADLINE = 10  # seconds


class WWTWebEngineView(QWebEngineView):
    # Pass drag and drop events back up to the parent
    # as this is needed for cases where applications
    # embed a PyWWT Qt widget.

    def dragEnterEvent(self, event):
        if self.parent() is None:
            super(WWTWebEngineView, self).dragEnterEvent(event)
        else:
            return self.parent().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        if self.parent() is None:
            super(WWTWebEngineView, self).dragMoveEvent(event)
        else:
            return self.parent().dragMoveEvent(event)

    def dragLeaveEvent(self, event):
        if self.parent() is None:
            super(WWTWebEngineView, self).dragLeaveEvent(event)
        else:
            return self.parent().dragLeaveEvent(event)

    def dropEvent(self, event):
        if self.parent() is None:
            super(WWTWebEngineView, self).dropEvent(event)
        else:
            return self.parent().dropEvent(event)


class WWTQWebEnginePage(QWebEnginePage):
    """
    Subclass of QWebEnginePage that abstracts the JavaScript invocation
    mechanism.
    """

    app_message_callback = None

    wwt_ready = QtCore.Signal()
    """A signal raised when the WWT app first becomes ready.

    The glue-wwt plugin requires this signal, so we can't move or remove it.
    """

    def __init__(self, parent=None):
        super(WWTQWebEnginePage, self).__init__(parent=parent)

        if WEBENGINE:
            self._js_response_received = False
            self._js_response = None
        else:
            self._frame = self.mainFrame()

    if WEBENGINE:

        def javaScriptConsoleMessage(
            self,
            level=None,
            message=None,
            line_number=None,
            source_id=None
        ):
            try:
                context = 'level={0}, line_number={1}, source_id={2}'.format(level, line_number, source_id)
                self._common_console_handler(message, context)
            except:  # noqa: E722
                logger.exception('unhandled Python exception in Qt webengine javaScriptConsoleMessage')  # noqa

        def _process_js_response(self, result):
            self._js_response_received = True
            self._js_response = result

        def runJavaScript(self, code):
            app = get_qapp()
            self._js_response_received = False
            self._js_response = None
            super(WWTQWebEnginePage, self).runJavaScript(code, self._process_js_response)

            while not self._js_response_received:
                app.processEvents()

            return self._js_response

    else:

        def javaScriptConsoleMessage(
            self,
            message=None,
            line_number=None,
            source_id=None
        ):
            try:
                context = 'line_number={0}, source_id={1}'.format(line_number, source_id)
                self._common_console_handler(message, context)
            except:  # noqa: E722
                logger.exception('unhandled Python exception in Qt webkit javaScriptConsoleMessage')  # noqa

        def runJavaScript(self, code):
            return self._frame.evaluateJavaScript(code)

    def _common_console_handler(self, message, context):
        if message.startswith('pywwtMessage:'):
            try:
                payload = json.loads(message[13:])
            except Exception as e:
                logger.warning('invalid pywwtMessage JSON: %s', e)
                return

            if self.app_message_callback is None:
                logger.warning('received app message, but no handler available; message: %s', payload)
            else:
                try:
                    self.app_message_callback(payload)
                except Exception:
                    logger.exception('error handling app message; payload: %s', payload)  # noqa
        else:
            logger.debug('JS console message: %s (%s)', message, context)


class WWTQtWidget(QtWidgets.QWidget):
    def __init__(self, url, parent=None):
        super(WWTQtWidget, self).__init__(parent=parent)

        self.web = WWTWebEngineView()
        self.page = WWTQWebEnginePage()
        self.page.setView(self.web)
        self.web.setPage(self.page)
        self.web.setUrl(QtCore.QUrl(url))

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        layout.addWidget(self.web)

    # More mouse-drag helpers

    def dragEnterEvent(self, event):
        if self.parent() is None:
            super(WWTQtWidget, self).dragEnterEvent(event)
        else:
            return self.parent().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        if self.parent() is None:
            super(WWTQtWidget, self).dragMoveEvent(event)
        else:
            return self.parent().dragMoveEvent(event)

    def dragLeaveEvent(self, event):
        if self.parent() is None:
            super(WWTQtWidget, self).dragLeaveEvent(event)
        else:
            return self.parent().dragLeaveEvent(event)

    def dropEvent(self, event):
        if self.parent() is None:
            super(WWTQtWidget, self).dropEvent(event)
        else:
            return self.parent().dropEvent(event)


class WWTQtClient(BaseWWTWidget):
    """
    A client to create and drive the Qt widget.

    Parameters
    ----------
    block_until_ready : `bool`
        Tells Python to wait for WorldWide Telescope to open before
        proceeding with any following script (default: `True`).

    size : `tuple`
        Sets size of widget in pixels (default: (600, 600)).

    hide_all_chrome : optional `bool`
        Configures the WWT frontend to hide all user-interface "chrome".
        Defaults to true to maintain compatibility with the historical
        pywwt user experience.
    """

    def __init__(self, block_until_ready=False, size=None, hide_all_chrome=True):
        app = get_qapp()

        self._data_server = get_data_server()

        wwt_url = self._data_server.static_url('qtwrapper.html')
        self.widget = WWTQtWidget(url=wwt_url)

        if size is not None:
            self.widget.resize(*size)

        self.widget.page.app_message_callback = self._on_app_message
        self.widget.show()

        super(WWTQtClient, self).__init__(
            hide_all_chrome=hide_all_chrome,
        )

        # Start polling for the app to start responding to messages

        self._last_pong_timestamp = 0
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self._check_ready)
        self._timer.start(1000)

        # TODO: this should be more generic
        if block_until_ready:
            while True:
                app.processEvents()
                if self._appAlive:
                    break

    def _check_ready(self):
        # If this Qt signal callback function raises an unhandled exception, it
        # can crash the whole process! Which is ridiculous but let's try to do
        # our part to make that not happen.
        #
        # Cf: https://doc.qt.io/qt-5/exceptionsafety.html#signals-and-slots
        try:
            self._check_ready_inner()
        except:  # noqa: E722
            logger.exception('unhandled exception in Qt check-ready callback')

    def _check_ready_inner(self):
        # Send the ping. We have some extra paranoia here in case funky
        # sequencing can happen in the stop() method.

        if self.widget is not None and self.widget.page is not None:
            self._actually_send_msg({
                'type': 'wwt_ping_pong',
                'sessionId': 'qt',
                'threadId': str(time.time()),
            })

        # Evaluate pong status, with a hack for glue-wwt -- we need to emit the
        # wwt_ready signal on our `page` field. It hardcodes the location of the
        # field so we can't rationalize this API.

        alive = (time.time() - self._last_pong_timestamp) < APP_LIVENESS_DEADLINE

        if not self._appAlive and alive:
            self.widget.page.wwt_ready.emit()

        self._on_app_status_change(alive=alive)

    def _on_app_message(self, payload):
        ptype = payload.get('type')

        if ptype == 'wwt_ping_pong':
            try:
                ts = float(payload['threadId'])
            except Exception:
                logger.exception('invalid timestamp in pywwt Qt pingpong response')
            else:
                self._last_pong_timestamp = ts
        else:
            self._on_app_message_received(payload)

    def wait(self, duration=None):
        """
        Prevents WorldWide Telescope from closing once Python reaches the
        end of a given script.

        Parameters
        ----------
        duration : int or float or None
            How many seconds to wait for. By default, this waits until the
            Qt window is closed.
        """
        app = get_qapp()

        if duration is None:
            app.exec_()
        else:
            time1 = time.time()
            while time.time() - time1 < duration:
                app.processEvents()

    def _actually_send_msg(self, payload):
        jmsg = json.dumps(payload)
        return self.widget.page.runJavaScript("pywwtSendMessage({0});".format(jmsg))

    def _serve_file(self, filename, extension=''):
        return self._data_server.serve_file(filename, extension=extension)

    def render(self, filename):
        """
        Saves a screenshot of the viewer's current state.

        Parameters
        ----------
        filename : `str`
            The desired name of the image file to be saved.
        """
        image = QtGui.QImage(self.widget.size(), QtGui.QImage.Format_RGB32)
        painter = QtGui.QPainter(image)
        self.widget.render(painter)
        image.save(filename)
        painter.end()

    def close(self):
        self._timer.stop()
        self.widget.page = None
        self.widget.web = None
        self.widget.close()
        self.widget = None



import asyncio
from hashlib import md5
import logging
import mimetypes
import os.path
import socket
from threading import Thread
import time

__all__ = ['get_data_server']

_data_server = None


def get_data_server(verbose=True):
    """
    This starts up a tornado server and returns a handle to a DataServer
    object which can be used to register files to serve.
    """

    global _data_server

    if _data_server is not None:
        return _data_server

    from tornado.ioloop import IOLoop
    from tornado.web import RequestHandler, Application, StaticFileHandler
    from tornado.routing import PathMatches

    class WebServer(Application):
        host = None
        port = None

        def run(self, host=None, port=8886):
            self.host = host
            self.port = port
            try:
                self.listen(port)
                IOLoop.instance().start()
            finally:
                self.host = None
                self.port = None

    class DataServer(object):
        def start(self, app):
            self._files = {}
            self._thread = Thread(target=self.start_app)
            self._thread.daemon = True
            self._thread.start()
            self._app = app
            while self._app.host is None and self._app.port is None:
                time.sleep(0.1)

        @property
        def port(self):
            return self._app.port

        @property
        def host(self):
            return self._app.host

        def start_app(self):

            asyncio.set_event_loop(asyncio.new_event_loop())

            host = socket.gethostbyname('localhost')

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(('localhost', 0))
            port = sock.getsockname()[1]
            sock.close()

            access_log = logging.getLogger("tornado.access")
            access_log.setLevel('ERROR')

            self._app.run(host=host, port=port)

        def serve_file(self, filename, extension=''):
            with open(filename, 'rb') as f:
                content = f.read()

            hash = md5(content).hexdigest() + extension
            self._files[hash] = os.path.abspath(filename)
            return 'http://' + self.host + ':' + str(self.port) + '/data/' + hash

        def static_url(self, rest):
            return 'http://' + self.host + ':' + str(self.port) + '/static/' + rest

        def get_file_contents(self, hash):
            with open(self._files[hash], 'rb') as f:
                return f.read()

    mimetypes.add_type('image/fits', '.fits')
    mimetypes.add_type('image/fits', '.fts')
    mimetypes.add_type('image/fits', '.fit')

    ds = DataServer()

    class DataHandler(RequestHandler):
        async def get(self, hash):
            # Do our best to set an appropriate Content-Type.
            filename = ds._files[hash]
            content_type = mimetypes.guess_type(filename)[0]
            if content_type is None:
                content_type = 'application/binary'
            self.set_header('Content-Type', content_type)

            # Add wide-open CORS headers to allow external WWT apps to access
            # data. This isn't needed in the default case, but comes in handy
            # when testing updates to the research app with an alternative
            # localhost port. Note that a hostile actor can just ignore these
            # settings, and our default stance is that data are globally
            # accessible, so this really shouldn't affect the level of security
            # we provide.
            self.set_header('Access-Control-Allow-Origin', '*')
            self.set_header('Access-Control-Allow-Methods', 'GET,HEAD')
            self.set_header('Access-Control-Allow-Headers', 'Content-Disposition,Content-Encoding,Content-Length,Content-Type')

            self.write(ds.get_file_contents(hash))

    app = WebServer([
        (PathMatches(r'/data/(?P<hash>\S+)'), DataHandler),
        (r'/static/(.*)', StaticFileHandler, {
            'path': os.path.join(os.path.dirname(__file__), 'web_static'),
            'default_filename': 'index.html',
        }),
    ])

    ds.start(app)

    _data_server = ds

    return ds



#! /usr/bin/env python
# -*- mode: python; coding: utf-8 -*-
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
import io
import os
import sys

from glob import glob
from os.path import join as pjoin

sys.path.append('.')
from setupbase import (
    create_cmdclass, install_npm, ensure_targets,
    find_packages, combine_commands, ensure_python,
    get_version, HERE
)

from setuptools import setup

ensure_python('>=3.7')

# Gather package metadata.

name = 'pywwt'
version = get_version(pjoin(name, '_version.py'))

with io.open('README.md', encoding='utf_8') as f:
    LONG_DESCRIPTION = f.read()

# Wire up the NPM build to the Python package build -- we generate JavaScript
# artifacts from the `frontend/` package and then include them as data assets in
# the pywwt Python package.

nb_path = pjoin(name, 'nbextension', 'static')

js_outputs = [
    pjoin(nb_path, 'index.js'),
]

js_content_command = combine_commands(
    install_npm('frontend', build_cmd='pywwt-export'),
    ensure_targets(js_outputs),
)

# Custom "command class" that (1) makes sure to create the JS content, (2)
# includes that content as extra "package data" in the Python package, and (3)
# can install special metadata files in the Python environment root.

package_data_spec = {
    name: [
        'interactive_figure/*.html',
        'interactive_figure/*.js',
        'labextension/*.tgz',
        'nbextension/static/*.*js*',
        'tests/data/*/*.png',
        'tests/data/*.fits',
        'web_static/**',
    ]
}

lab_path = pjoin(name, 'labextension')
local_nbdotd_path = pjoin('jupyter.d', 'notebook.d')
local_jnbcfgdotd_path = pjoin('jupyter.d', 'jupyter_notebook_config.d')

data_files_spec = [
    ('share/jupyter/nbextensions/pywwt', nb_path, '*.js*'),
    ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ('etc/jupyter/nbconfig/notebook.d', local_nbdotd_path, 'pywwt.json'),
    ('etc/jupyter/jupyter_notebook_config.d' ,local_jnbcfgdotd_path, 'pywwt.json'),
]

cmdclass = create_cmdclass(
    'js-content',
    package_data_spec = package_data_spec,
    data_files_spec = data_files_spec
)
cmdclass['js-content'] = js_content_command

# Ready to go!

setup_args = dict(
    name            = name,
    description     = 'The AAS WorldWide Telescope from Python',
    long_description = LONG_DESCRIPTION,
    long_description_content_type = 'text/markdown',
    version         = version,
    cmdclass        = cmdclass,
    packages        = find_packages(),
    author          = 'Thomas P. Robitaille, O. Justin Otor, and John ZuHone',
    author_email    = 'thomas.robitaille@gmail.com',
    url             = 'https://github.com/WorldWideTelescope/pywwt',
    license         = 'BSD',
    platforms       = "Linux, Mac OS X, Windows",
    keywords        = ['Jupyter', 'Widgets', 'IPython'],
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Jupyter',
    ],
    include_package_data = True,
    python_requires = '>=3.7',
    install_requires = [
        'numpy>=1.9',
        'matplotlib>1.5',
        'astropy>=1.0,!=4.0.1',
        'requests',
        'beautifulsoup4',
        'python-dateutil',
        'lxml',
        'ipywidgets>=7.0.0',
        'ipyevents',
        'traitlets>=5',
        'reproject>=0.8',
        'qtpy',
        'tornado',
        'pytz',
    ],
    extras_require = {
        'test': [
            'pytest',
            'pytest-cov>=2.6.1',
            'pytest-remotedata>=0.3.1',
        ],
        'docs': [
            'sphinx>=1.6',
            'sphinx-automodapi',
            'numpydoc',
            'sphinx_rtd_theme',
            'jupyter_sphinx',
        ],
        'qt': [
            'PyQt5;python_version>="3"',
            'PyQtWebEngine;python_version>="3"',
        ],
        'lab': [
            'jupyterlab'
        ]
    },
    entry_points = {
    },
)

if __name__ == '__main__':
    setup(**setup_args)
